@@ -82,9 +82,9 @@ struct _LmConnection {
 	LmCallback   *open_cb;
 
 	gboolean      cancel_open;
-	LmCallback   *close_cb;
+	LmCallback   *close_cb;     /* unused */
 	LmCallback   *auth_cb;
-	LmCallback   *register_cb;
+	LmCallback   *register_cb;  /* unused */
 
 	LmCallback   *disconnect_cb;
 
@@ -96,6 +96,9 @@ struct _LmConnection {
 	guint         keep_alive_rate;
 	guint         keep_alive_id;
 
+	guint         io_watch_out;
+	GString      *out_buf;
+
 	gint          ref_count;
 };
 
@@ -118,6 +121,10 @@ static gboolean connection_do_open           (LmConnection         *connection,
 					      GError              **error);
 
 static void     connection_do_close           (LmConnection        *connection);
+static gint     connection_do_write          (LmConnection         *connection,
+					      const gchar          *buf,
+					      gint                  len);
+
 static gboolean connection_in_event          (GIOChannel   *source,
 					      GIOCondition  condition,
 					      LmConnection *connection);
@@ -127,7 +134,7 @@ static gboolean connection_error_event       (GIOChannel   *source,
 static gboolean connection_hup_event         (GIOChannel   *source,
 					      GIOCondition  condition,
 					      LmConnection *connection);
-static gboolean connection_send              (LmConnection             *connection,
+static gboolean connection_send              (LmConnection         *connection,
 					      const gchar          *str,
 					      gint                  len,
 					      GError               **error);
@@ -148,8 +155,8 @@ static LmHandlerResult connection_auth_reply (LmMessageHandler    *handler,
 					      LmMessage           *m,
 					      gpointer             user_data);
 
-static void     connection_stream_received   (LmConnection             *connection, 
-					      LmMessage                *m);
+static void     connection_stream_received   (LmConnection        *connection, 
+					      LmMessage           *m);
 
 static gint     connection_handler_compare_func (HandlerData  *a,
 						 HandlerData  *b);
@@ -172,6 +179,15 @@ static guint    connection_add_watch            (LmConnection  *connection,
 static gboolean connection_send_keep_alive      (LmConnection  *connection);
 static void     connection_start_keep_alive     (LmConnection  *connection);
 static void     connection_stop_keep_alive      (LmConnection  *connection);
+static gboolean connection_buffered_write_cb    (GIOChannel    *source, 
+						 GIOCondition   condition,
+						 LmConnection  *connection);
+static gboolean connection_output_is_buffered   (LmConnection  *connection,
+						 const gchar   *buffer,
+						 gint           len);
+static void     connection_setup_output_buffer  (LmConnection  *connection,
+						 const gchar   *buffer,
+						 gint           len);
 
 static GSourceFuncs incoming_funcs = {
 	connection_incoming_prepare,
@@ -189,6 +205,10 @@ connection_free (LmConnection *connection)
 	g_free (connection->server);
 	g_free (connection->jid);
 
+	if (connection->parser) {
+		lm_parser_free (connection->parser);
+	}
+
 	/* Unref handlers */
 	for (i = 0; i < LM_MESSAGE_TYPE_UNKNOWN; ++i) {
 		GSList *l;
@@ -208,6 +228,16 @@ connection_free (LmConnection *connection)
 		connection_do_close (connection);
 	}
 
+	if (connection->open_cb) {
+		_lm_utils_free_callback (connection->open_cb);
+	}
+	
+	if (connection->auth_cb) {
+		_lm_utils_free_callback (connection->auth_cb);
+	}
+
+	lm_connection_set_disconnect_function (connection, NULL, NULL, NULL);
+
 	while ((m = g_queue_pop_head (connection->incoming_messages)) != NULL) {
 		lm_message_unref (m);
 	}
@@ -221,6 +251,10 @@ connection_free (LmConnection *connection)
         if (connection->context) {
                 g_main_context_unref (connection->context);
         }
+
+	if (connection->out_buf) {
+		g_string_free (connection->out_buf, TRUE);
+	}
         
         g_free (connection);
 }
@@ -233,9 +267,11 @@ connection_handle_message (LmConnection *connection, LmMessage *m)
 	const gchar      *id;
 	LmHandlerResult   result = LM_HANDLER_RESULT_ALLOW_MORE_HANDLERS;
 
+	lm_connection_ref (connection);
+
 	if (lm_message_get_type (m) == LM_MESSAGE_TYPE_STREAM) {
 		connection_stream_received (connection, m);
-		return;
+		goto out;
 	}
 	
 	id = lm_message_node_get_attribute (m->node, "id");
@@ -251,7 +287,7 @@ connection_handle_message (LmConnection *connection, LmMessage *m)
 	}
 	
 	if (result == LM_HANDLER_RESULT_REMOVE_MESSAGE) {
-		return;
+		goto out;
 	}
 
 	for (l = connection->handlers[lm_message_get_type (m)]; 
@@ -263,6 +299,9 @@ connection_handle_message (LmConnection *connection, LmMessage *m)
 							     connection,
 							     m);
 	}
+
+out:
+	lm_connection_unref (connection);
 	
 	return;
 }
@@ -365,7 +404,9 @@ _lm_connection_succeeded (LmConnectData *connect_data)
 				      G_IO_HUP,
 				      (GIOFunc) connection_hup_event,
 				      connection);
-	
+
+	/* FIXME: Set up according to XMPP 1.0 specification */
+	/*        StartTLS and the like */
 	if (!connection_send (connection, 
 			      "<?xml version='1.0' encoding='UTF-8'?>", -1,
 			      NULL)) {
@@ -424,12 +465,16 @@ _lm_connection_failed_with_error (LmConnectData *connect_data, int error)
 	}
 	
 	if (connect_data->current_addr == NULL) {
-		connection->state = LM_CONNECTION_STATE_CLOSED;
-		if (connection->open_cb && connection->open_cb->func) {
-			LmCallback *cb = connection->open_cb;
+		connection_do_close (connection);
+		if (connection->open_cb) {
+			LmCallback *cb;
+
+			cb = connection->open_cb;
+			connection->open_cb = NULL;
 			
 			(* ((LmResultFunction) cb->func)) (connection, FALSE,
 							   cb->user_data);
+			_lm_utils_free_callback (cb);
 		}
 		
 		freeaddrinfo (connect_data->resolved_addrs);
@@ -445,7 +490,7 @@ _lm_connection_failed (LmConnectData *connect_data)
 {
 	_lm_connection_failed_with_error (connect_data,errno);
 }
-	
+
 static gboolean 
 connection_connect_cb (GIOChannel   *source, 
 		       GIOCondition  condition,
@@ -603,6 +648,75 @@ connection_stop_keep_alive (LmConnection *connection)
 	connection->keep_alive_id = 0;
 }
 
+static gboolean
+connection_buffered_write_cb (GIOChannel   *source, 
+			      GIOCondition  condition,
+			      LmConnection *connection)
+{
+	gint     b_written;
+	GString *out_buf;
+	/* FIXME: Do the writing */
+
+	out_buf = connection->out_buf;
+	if (!out_buf) {
+		/* Should not be possible */
+		return FALSE;
+	}
+
+	b_written = connection_do_write (connection, out_buf->str, out_buf->len);
+
+	if (b_written < 0) {
+		connection_error_event (connection->io_channel, 
+					G_IO_HUP,
+					connection);
+		return FALSE;
+	}
+
+	g_string_erase (out_buf, 0, (gsize) b_written);
+	if (out_buf->len == 0) {
+		lm_verbose ("Output buffer is empty, going back to normal output\n");
+		g_source_destroy (g_main_context_find_source_by_id (
+			connection->context, connection->io_watch_out));
+		connection->io_watch_out = 0;
+		g_string_free (out_buf, TRUE);
+		connection->out_buf = NULL;
+		return FALSE;
+	}
+
+	return TRUE;
+}
+
+static gboolean
+connection_output_is_buffered (LmConnection *connection,
+			       const gchar  *buffer,
+			       gint          len)
+{
+	if (connection->out_buf) {
+		lm_verbose ("Appending %d bytes to output buffer\n", len);
+		g_string_append_len (connection->out_buf, buffer, len);
+		return TRUE;
+	}
+
+	return FALSE;
+}
+
+static void
+connection_setup_output_buffer (LmConnection *connection,
+				const gchar  *buffer,
+				gint          len)
+{
+	lm_verbose ("OUTPUT BUFFER ENABLED\n");
+
+	connection->out_buf = g_string_new_len (buffer, len);
+
+	connection->io_watch_out =
+		connection_add_watch (connection,
+				      connection->io_channel,
+				      G_IO_OUT,
+				      (GIOFunc) connection_buffered_write_cb,
+				      connection);
+}
+
 /* Returns directly */
 /* Setups all data needed to start the connection attempts */
 static gboolean
@@ -615,7 +729,7 @@ connection_do_open (LmConnection *connection, GError **error)
 	if (lm_connection_is_open (connection)) {
 		g_set_error (error,
 			     LM_ERROR,
-			     LM_ERROR_CONNECTION_NOT_OPEN,
+			     LM_ERROR_CONNECTION_OPEN,
 			     "Connection is already open, call lm_connection_close() first");
 		return FALSE;
 	}
@@ -623,7 +737,7 @@ connection_do_open (LmConnection *connection, GError **error)
 	if (!connection->server) {
 		g_set_error (error,
 			     LM_ERROR,
-			     LM_ERROR_CONNECTION_OPEN,
+			     LM_ERROR_CONNECTION_FAILED,
 			     "You need to set the server hostname in the call to lm_connection_new()");
 		return FALSE;
 	}
@@ -655,7 +769,7 @@ connection_do_open (LmConnection *connection, GError **error)
 		if (getaddrinfo (proxy_server, NULL, &req, &ans) != 0) {
 			g_set_error (error,
 				     LM_ERROR,                 
-				     LM_ERROR_CONNECTION_OPEN,   
+				     LM_ERROR_CONNECTION_FAILED,   
 				     "getaddrinfo() failed");
 			return FALSE;
 		}
@@ -668,7 +782,7 @@ connection_do_open (LmConnection *connection, GError **error)
 				 NULL, &req, &ans) != 0) {
 			g_set_error (error,
 				     LM_ERROR,                 
-				     LM_ERROR_CONNECTION_OPEN,   
+				     LM_ERROR_CONNECTION_FAILED,   
 				     "getaddrinfo() failed");
 			return FALSE;
 		}
@@ -704,6 +818,12 @@ connection_do_close (LmConnection *connection)
 		g_source_destroy (g_main_context_find_source_by_id (
 			connection->context, connection->io_watch_hup));
 
+		if (connection->io_watch_out != 0) {
+			g_source_destroy (g_main_context_find_source_by_id (
+				connection->context, connection->io_watch_out));
+			connection->io_watch_out = 0;
+		}
+
 		if (connection->io_watch_connect != 0) {
 			g_source_destroy (g_main_context_find_source_by_id(connection->context,
 									   connection->io_watch_connect));
@@ -715,22 +835,51 @@ connection_do_close (LmConnection *connection)
 
 		connection->fd = -1;
 	}
-
 	
 	g_source_destroy (connection->incoming_source);
 	g_source_unref (connection->incoming_source);
 
 	if (!lm_connection_is_open (connection)) {
+		/* lm_connection_is_open is FALSE for state OPENING as well */
+		connection->state = LM_CONNECTION_STATE_CLOSED;
 		return;
 	}
 	
 	connection->state = LM_CONNECTION_STATE_CLOSED;
-
 	if (connection->ssl) {
 		_lm_ssl_close (connection->ssl);
 	}
 }
 
+static gint
+connection_do_write (LmConnection *connection,
+		     const gchar  *buf,
+		     gint          len)
+{
+	gint b_written;
+
+	if (connection->ssl) {
+		b_written = _lm_ssl_send (connection->ssl, buf, len);
+	} else {
+		GIOStatus io_status = G_IO_STATUS_AGAIN;
+		gsize     bytes_written;
+
+		while (io_status == G_IO_STATUS_AGAIN) {
+			io_status = g_io_channel_write_chars (connection->io_channel, 
+							      buf, len, 
+							      &bytes_written,
+							      NULL);
+		}
+
+		b_written = bytes_written;
+
+		if (io_status != G_IO_STATUS_NORMAL) {
+			b_written = -1;
+		}
+	}
+
+	return b_written;
+}
 
 static gboolean
 connection_in_event (GIOChannel   *source,
@@ -834,7 +983,7 @@ connection_send (LmConnection  *connection,
 		 gint           len, 
 		 GError       **error)
 {
-	gsize             bytes_written;
+	gint b_written;
 	
 	if (connection->state < LM_CONNECTION_STATE_OPENING) {
 		g_set_error (error,
@@ -855,16 +1004,27 @@ connection_send (LmConnection  *connection,
 	g_log (LM_LOG_DOMAIN, LM_LOG_LEVEL_NET, 
 	       "-----------------------------------\n");
 
-	if (connection->ssl) {
-		if (!_lm_ssl_send (connection->ssl, str, len)) {
-			
-			connection_error_event (connection->io_channel, 
-						G_IO_HUP,
-						connection);
-		}
-	} else {
-		g_io_channel_write_chars (connection->io_channel, str, len, 
-					  &bytes_written, NULL);
+	/* Check to see if there already is an output buffer, if so, add to the
+	   buffer and return */
+
+	if (connection_output_is_buffered (connection, str, len)) {
+		return TRUE;
+	}
+
+	b_written = connection_do_write (connection, str, len);
+
+
+	if (b_written < 0) {
+		connection_error_event (connection->io_channel, 
+					G_IO_HUP,
+					connection);
+		return FALSE;
+	}
+
+	if (b_written < len) {
+		connection_setup_output_buffer (connection, 
+						str + b_written, 
+						len - b_written);
 	}
 
 	return TRUE;
@@ -1032,16 +1192,20 @@ connection_auth_reply (LmMessageHandler *handler,
 	
 	lm_verbose ("AUTH reply: %d\n", result);
 	
-	if (connection->auth_cb && connection->auth_cb->func) {
-		LmCallback *cb = connection->auth_cb;
+	if (connection->auth_cb) {
+	        LmCallback *cb;
 
-		(* ((LmResultFunction) cb->func)) (connection, 
-						   result, cb->user_data);
+		cb = connection->auth_cb;
+		connection->auth_cb = NULL;
+
+		if (cb->func) {
+	    		(* ((LmResultFunction) cb->func)) (connection, 
+						           result, cb->user_data);
+		}
+
+		_lm_utils_free_callback (cb);
 	}
 	
-	_lm_utils_free_callback (connection->auth_cb);
-	connection->auth_cb = NULL;
-	
 	return LM_HANDLER_RESULT_REMOVE_MESSAGE;
 }
 
@@ -1066,15 +1230,19 @@ connection_stream_received (LmConnection *connection, LmMessage *m)
 
 	connection_start_keep_alive (connection);
 
-	if (connection->open_cb && connection->open_cb->func) {
-		LmCallback *cb = connection->open_cb;
+	if (connection->open_cb) {
+		LmCallback *cb;
+
+		cb = connection->open_cb;
+		connection->open_cb = NULL;
 		
-		(* ((LmResultFunction) cb->func)) (connection, result,
-						   cb->user_data);
+		if (cb->func) {
+		        (* ((LmResultFunction) cb->func)) (connection, result,
+			        			   cb->user_data);
+		}
+
+		_lm_utils_free_callback (connection->open_cb);
 	}
-	
-	_lm_utils_free_callback (connection->open_cb);
-	connection->open_cb = NULL;
 }
 
 static gint
@@ -1179,6 +1347,7 @@ lm_connection_new (const gchar *server)
 	connection->state             = LM_CONNECTION_STATE_CLOSED;
 	connection->keep_alive_id     = 0;
 	connection->keep_alive_rate   = 0;
+	connection->out_buf           = NULL;
 	
 	connection->id_handlers = g_hash_table_new_full (g_str_hash, 
 							 g_str_equal,
@@ -1285,6 +1454,12 @@ lm_connection_open_and_block (LmConnection *connection, GError **error)
 		return TRUE;
 	}
 
+	/* Need to set the error here: LM-15 */
+	g_set_error (error,
+		     LM_ERROR,
+		     LM_ERROR_CONNECTION_FAILED,
+		     "Opening the connection failed");
+
 	return FALSE;
 }
 
@@ -1315,9 +1490,11 @@ gboolean
 lm_connection_close (LmConnection      *connection, 
 		     GError           **error)
 {
+	gboolean no_errors = TRUE;
+
 	g_return_val_if_fail (connection != NULL, FALSE);
-	
-	if (!lm_connection_is_open (connection)) {
+
+	if (connection->state == LM_CONNECTION_STATE_CLOSED) {
 		g_set_error (error,
 			     LM_ERROR,
 			     LM_ERROR_CONNECTION_NOT_OPEN,
@@ -1328,16 +1505,18 @@ lm_connection_close (LmConnection      *connection,
 	lm_verbose ("Disconnecting from: %s:%d\n", 
 		    connection->server, connection->port);
 	
-	if (!connection_send (connection, "</stream:stream>", -1, error)) {
-		return FALSE;
+	if (lm_connection_is_open (connection)) {
+		if (!connection_send (connection, "</stream:stream>", -1, error)) {
+			no_errors = FALSE;
+		}
+		
+		g_io_channel_flush (connection->io_channel, NULL);
 	}
 	
- 	g_io_channel_flush (connection->io_channel, NULL);
-	
 	connection_do_close (connection);
 	connection_signal_disconnect (connection, LM_DISCONNECT_REASON_OK);
 	
-	return TRUE;
+	return no_errors;
 }
 
 /**
@@ -1383,6 +1562,8 @@ lm_connection_authenticate (LmConnection      *connection,
 		return FALSE;
 	}
 
+	/* FIXME: Do SASL authentication here (if XMPP 1.0 is used) */
+
 	connection->state = LM_CONNECTION_STATE_AUTHENTICATING;
 	
 	connection->auth_cb = _lm_utils_new_callback (function, 
@@ -1983,9 +2164,13 @@ lm_connection_set_disconnect_function (LmConnection         *connection,
 		_lm_utils_free_callback (connection->disconnect_cb);
 	}
 		
-	connection->disconnect_cb = _lm_utils_new_callback (function, 
-							    user_data,
-							    notify);
+	if (function) {
+		connection->disconnect_cb = _lm_utils_new_callback (function, 
+	    					                    user_data,
+							            notify);
+	} else {
+		connection->disconnect_cb = NULL;
+	}
 }
 
 /**
